# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""BGP Update Message"""

import struct
import traceback
import logging

import netaddr

from yabgp.common import exception as excep
from yabgp.common import constants as bgp_cons
from yabgp.message.attribute import AttributeFlag
from yabgp.message.attribute.origin import Origin
from yabgp.message.attribute.aspath import ASPath
from yabgp.message.attribute.nexthop import NextHop
from yabgp.message.attribute.med import MED
from yabgp.message.attribute.localpref import LocalPreference
from yabgp.message.attribute.atomicaggregate import AtomicAggregate
from yabgp.message.attribute.aggregator import Aggregator
from yabgp.message.attribute.community import Community
from yabgp.message.attribute.originatorid import OriginatorID
from yabgp.message.attribute.clusterlist import ClusterList

LOG = logging.getLogger()


class Update(object):

    """
    An UPDATE message is used to advertise feasible routes that share
    common path attributes to a peer, or to withdraw multiple unfeasible
    routes from service (RFC 4271 page 15)
    """

    def __init__(self):
        """
        +----------------------------------------------------+
        |     Withdrawn Routes Length (2 octets)             |
        +----------------------------------------------------+
        |     Withdrawn Routes (variable)                    |
        +----------------------------------------------------+
        |     Total Path Attribute Length (2 octets)         |
        +----------------------------------------------------+
        |     Path Attributes (variable)                     |
        +----------------------------------------------------+
        |  Network Layer Reachability Information (variable) |
        +----------------------------------------------------+

        @ Withdrawn Routes Length:
            This 2-octets unsigned integer indicates the total length of
        the Withdrawn Routes field in octets. Its value allows the
        length of the Network Layer Reachability Information field to
        be determined, as specified below.
            A value of 0 indicates that no routes are being withdrawn from
        service, and that the WITHDRAWN ROUTES field is not present in
        this UPDATE message.

        @ Withdrawn Routes:
            This is a variable-length field that contains a list of IP
        address prefixes for the routes that are being withdrawn from
        service. Each IP address prefix is encoded as a 2-tuple of the
        form <length, prefix>, whose fields are described below:
                    +---------------------------+
                    |     Length (1 octet)      |
                    +---------------------------+
                    |     Prefix (variable)     |
                    +---------------------------+
            The use and the meaning of these fields are as follows:
            a) Length:
            The Length field indicates the length in bits of the IP
            address prefix. A length of zero indicates a prefix that
            matches all IP addresses (with prefix, itself, of zero
            octets).
            b) Prefix:
            The Prefix field contains an IP address prefix, followed by
            the minimum number of trailing bits needed to make the end
            of the field fall on an octet boundary. Note that the value
            of trailing bits is irrelevant.

        @ Total Path Attribute Length:
            This 2-octet unsigned integer indicates the total length of the
        Path Attributes field in octets. Its value allows the length
        of the Network Layer Reachability field to be determined as
        specified below.
            A value of 0 indicates that neither the Network Layer
        Reachability Information field nor the Path Attribute field is
        present in this UPDATE message.

        @ Path Attributes:
            (path attributes details see RFC 4271 and some other RFCs)

        @ Network Layer Reachability Information:
            This variable length field contains a list of IP address
        prefixes. The length, in octets, of the Network Layer
        Reachability Information is not encoded explicitly, but can be
        calculated as:
        UPDATE message Length - 23 - Total Path Attributes Length
        - Withdrawn Routes Length
        where UPDATE message Length is the value encoded in the fixedsize
        BGP header, Total Path Attribute Length, and Withdrawn
        Routes Length are the values encoded in the variable part of
        the UPDATE message, and 23 is a combined length of the fixedsize
        BGP header, the Total Path Attribute Length field, and the
        Withdrawn Routes Length field.
        Reachability information is encoded as one or more 2-tuples of
        the form <length, prefix>, whose fields are described below:
                    +---------------------------+
                    |      Length (1 octet)     |
                    +---------------------------+
                    |      Prefix (variable)    |
                    +---------------------------+
        The use and the meaning of these fields are as follows:
            a) Length:
                The Length field indicates the length in bits of the IP
            address prefix. A length of zero indicates a prefix that
            matches all IP addresses (with prefix, itself, of zero
            octets).
            b) Prefix:
                The Prefix field contains an IP address prefix, followed by
            enough trailing bits to make the end of the field fall on an
            octet boundary. Note that the value of the trailing bits is
            irrelevant.
        """

    @classmethod
    def parse(cls, msg):

        """
        Parse BGP Update message

        :param msg: raw BGP update binary PDU data
        :type msg: list
        :return: message after parsing.
        :return type: dict

        """
        t = msg[0]
        asn4 = msg[1]
        msg_hex = msg[2]

        results = {
            "withdraw": None,
            "attr": None,
            "nlri": None,
            'time': t,
            'hex': msg_hex,
            'sub_error': None,
            'err_data': None}

        # get every part of the update message
        withdraw_len = struct.unpack('!H', msg_hex[:2])[0]
        withdraw_prefix_data = msg_hex[2:withdraw_len + 2]
        attr_len = struct.unpack('!H', msg_hex[withdraw_len + 2:withdraw_len + 4])[0]
        attribute_data = msg_hex[withdraw_len + 4:withdraw_len + 4 + attr_len]
        nlri_data = msg_hex[withdraw_len + 4 + attr_len:]

        try:
            # parse withdraw prefixes
            results['withdraw'] = cls.parse_prefix_list(withdraw_prefix_data)

            # parse nlri
            results['nlri'] = cls.parse_prefix_list(nlri_data)
        except Exception as e:
            LOG.error(e)
            error_str = traceback.format_exc()
            LOG.debug(error_str)
            results['sub_error'] = bgp_cons.ERR_MSG_UPDATE_INVALID_NETWORK_FIELD
            results['err_data'] = ''
        try:
            # parse attributes
            results['attr'] = cls.parse_attributes(attribute_data, asn4)
        except excep.UpdateMessageError as e:
            LOG.error(e)
            results['sub_error'] = e.sub_error
            results['err_data'] = e.data
        except Exception as e:
            LOG.error(e)
            error_str = traceback.format_exc()
            LOG.debug(error_str)
            results['sub_error'] = e
            results['err_data'] = e

        return results

    @classmethod
    def construct(cls, msg_dict, asn4=False):
        """construct BGP update message

        :param msg_dict: update message string
        :param asn4: support 4 bytes asn or not"""
        attr_hex = b''
        nlri_hex = b''
        withdraw_hex = b''
        if msg_dict['attr']:
            attr_hex = cls.construct_attributes(msg_dict['attr'], asn4)
        if msg_dict['nlri']:
            nlri_hex = cls.construct_prefix_v4(msg_dict['nlri'])
        if msg_dict['withdraw']:
            withdraw_hex = cls.construct_prefix_v4(msg_dict['withdraw'])
        if nlri_hex and attr_hex:
            msg_body = struct.pack('!H', 0) + struct.pack('!H', len(attr_hex)) + attr_hex + nlri_hex
            return cls.construct_header(msg_body)
        elif attr_hex and not nlri_hex:
            msg_body = struct.pack('!H', 0) + struct.pack('!H', len(attr_hex)) + attr_hex + nlri_hex
            return cls.construct_header(msg_body)
        elif withdraw_hex:
            msg_body = struct.pack('!H', len(withdraw_hex)) + withdraw_hex + struct.pack('!H', 0)
            return cls.construct_header(msg_body)

    @staticmethod
    def parse_prefix_list(data):
        """
        Parses an RFC4271 encoded blob of BGP prefixes into a list

        :param data:
        :return: prefix_list
        """
        prefixes = []
        postfix = data
        while len(postfix) > 0:
            # for python2 and python3
            if isinstance(postfix[0], int):
                prefix_len = postfix[0]
            else:
                prefix_len = ord(postfix[0])
            if prefix_len > 32:
                LOG.warning('Prefix Length larger than 32')
                raise excep.UpdateMessageError(
                    sub_error=bgp_cons.ERR_MSG_UPDATE_INVALID_NETWORK_FIELD,
                    data=repr(data)
                )
            octet_len, remainder = int(prefix_len / 8), prefix_len % 8
            if remainder > 0:
                # prefix length doesn't fall on octet boundary
                octet_len += 1
            tmp = postfix[1:octet_len + 1]
            # for python2 and python3
            if isinstance(postfix[0], int):
                prefix_data = [i for i in tmp]
            else:
                prefix_data = [ord(i) for i in tmp]
            # Zero the remaining bits in the last octet if it didn't fall
            # on an octet boundary
            if remainder > 0:
                prefix_data[-1] &= 255 << (8 - remainder)
            prefix_data = prefix_data + list(str(0)) * 4
            prefix = "%s.%s.%s.%s" % (tuple(prefix_data[0:4])) + '/' + str(prefix_len)
            prefixes.append(prefix)
            # Next prefix
            postfix = postfix[octet_len + 1:]

        return prefixes

    @staticmethod
    def parse_attributes(data, asn4=False):
        """
        Parses an RFC4271 encoded blob of BGP attributes into a list

        :param data:
        :param asn4: support 4 bytes asn or not
        :return:
        """
        attributes = {}
        postfix = data
        while len(postfix) > 0:

            try:
                flags, type_code = struct.unpack('!BB', postfix[:2])

                if flags & AttributeFlag.EXTENDED_LENGTH:
                    attr_len = struct.unpack('!H', postfix[2:4])[0]
                    attr_value = postfix[4:4 + attr_len]
                    postfix = postfix[4 + attr_len:]    # Next attribute
                else:    # standard 1-octet length
                    if isinstance(postfix[2], int):
                        attr_len = postfix[2]
                    else:
                        attr_len = ord(postfix[2])
                    attr_value = postfix[3:3 + attr_len]
                    postfix = postfix[3 + attr_len:]    # Next attribute
            except Exception as e:
                LOG.error(e)
                error_str = traceback.format_exc()
                LOG.debug(error_str)
                raise excep.UpdateMessageError(
                    sub_error=bgp_cons.ERR_MSG_UPDATE_MALFORMED_ATTR_LIST,
                    data='')

            if type_code == bgp_cons.BGPTYPE_ORIGIN:

                decode_value = Origin.parse(value=attr_value)

            elif type_code == bgp_cons.BGPTYPE_AS_PATH:

                decode_value = ASPath.parse(value=attr_value, asn4=asn4)

            elif type_code == bgp_cons.BGPTYPE_NEXT_HOP:

                decode_value = NextHop.parse(value=attr_value)

            elif type_code == bgp_cons.BGPTYPE_MULTI_EXIT_DISC:

                decode_value = MED.parse(value=attr_value)

            elif type_code == bgp_cons.BGPTYPE_LOCAL_PREF:

                decode_value = LocalPreference.parse(value=attr_value)

            elif type_code == bgp_cons.BGPTYPE_ATOMIC_AGGREGATE:

                decode_value = AtomicAggregate.parse(value=attr_value)

            elif type_code == bgp_cons.BGPTYPE_AGGREGATOR:

                decode_value = Aggregator.parse(value=attr_value, asn4=asn4)

            elif type_code == bgp_cons.BGPTYPE_COMMUNITIES:

                decode_value = Community.parse(value=attr_value)

            elif type_code == bgp_cons.BGPTYPE_ORIGINATOR_ID:

                decode_value = OriginatorID.parse(value=attr_value)

            elif type_code == bgp_cons.BGPTYPE_CLUSTER_LIST:

                decode_value = ClusterList.parse(value=attr_value)

            elif type_code == bgp_cons.BGPTYPE_NEW_AS_PATH:

                decode_value = ASPath.parse(value=attr_value, asn4=True)

            elif type_code == bgp_cons.BGPTYPE_NEW_AGGREGATOR:

                decode_value = Aggregator.parse(value=attr_value, asn4=True)
            else:
                decode_value = repr(attr_value)
            attributes[type_code] = decode_value

        return attributes

    @staticmethod
    def construct_attributes(attr_dict, asn4=False):

        """
        construts BGP Update attirubte.

        :param attr_dict: bgp attribute dictionary
        :param asn4: support 4 bytes asn or not
        """
        attr_raw_hex = b''
        for type_code, value in attr_dict.items():

            if type_code == bgp_cons.BGPTYPE_ORIGIN:
                origin_hex = Origin.construct(value=value)
                attr_raw_hex += origin_hex

            elif type_code == bgp_cons.BGPTYPE_AS_PATH:
                aspath_hex = ASPath.construct(value=value, asn4=asn4)
                attr_raw_hex += aspath_hex

            elif type_code == bgp_cons.BGPTYPE_NEXT_HOP:
                nexthop_hex = NextHop.construct(value=value)
                attr_raw_hex += nexthop_hex

            elif type_code == bgp_cons.BGPTYPE_MULTI_EXIT_DISC:
                med_hex = MED.construct(value=value)
                attr_raw_hex += med_hex

            elif type_code == bgp_cons.BGPTYPE_LOCAL_PREF:
                localpre_hex = LocalPreference.construct(value=value)
                attr_raw_hex += localpre_hex

            elif type_code == bgp_cons.BGPTYPE_ATOMIC_AGGREGATE:
                atomicaggregate_hex = AtomicAggregate.construct(value=value)
                attr_raw_hex += atomicaggregate_hex

            elif type_code == bgp_cons.BGPTYPE_AGGREGATOR:
                aggregator_hex = Aggregator.construct(value=value, asn4=asn4)
                attr_raw_hex += aggregator_hex

            elif type_code == bgp_cons.BGPTYPE_COMMUNITIES:
                community_hex = Community.construct(value=value)
                attr_raw_hex += community_hex

            elif type_code == bgp_cons.BGPTYPE_ORIGINATOR_ID:
                originatorid_hex = OriginatorID.construct(value=value)
                attr_raw_hex += originatorid_hex

            elif type_code == bgp_cons.BGPTYPE_CLUSTER_LIST:
                clusterlist_hex = ClusterList.construct(value=value)
                attr_raw_hex += clusterlist_hex

        return attr_raw_hex

    @staticmethod
    def construct_header(msg):
        """
        Prepends the mandatory header to a constructed BGP message

        :param msg:
        :return:
        """
        #    16-octet     2-octet  1-octet
        # ---------------+--------+---------+------+
        #    Maker      | Length |  Type   |  msg |
        # ---------------+--------+---------+------+
        return b'\xff'*16 + struct.pack('!HB', len(msg) + 19, 2) + msg

    @staticmethod
    def construct_prefix_v4(prefix_list):
        """
        constructs NLRI prefix list

        :param prefix_list: prefix list
        """
        nlri_raw_hex = b''
        for prefix in prefix_list:
            masklen = prefix.split('/')[1]
            ip_hex = struct.pack('!I', netaddr.IPNetwork(prefix).value)
            masklen = int(masklen)
            if 16 < masklen <= 24:
                ip_hex = ip_hex[0:3]
            elif 8 < masklen <= 16:
                ip_hex = ip_hex[0:2]
            elif masklen <= 8:
                ip_hex = ip_hex[0:1]
            nlri_raw_hex += struct.pack('!B', masklen) + ip_hex
        return nlri_raw_hex



# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

""" BGP Open message"""

import struct
import netaddr

from yabgp.common import exception as excp
from yabgp.common import constants as bgp_cons


class Open(object):
    """
    After a TCP connection is established, the first message sent by each
    side is an OPEN message. If the OPEN message is acceptable, a
    KEEPALIVE message confirming the OPEN is sent back
    """

    def __init__(self, version=None, asn=None, hold_time=None,
                 bgp_id=None, opt_para_len=None, opt_paras=None):

        """
        :param version: BGP Protocol version.
        :param asn: AS number.
        :param hold_time: Hold time
        :param bgp_id: BGP Router ID
        :param opt_para_len: Optional Parameters length
        :param opt_paras: Optional Parameters
        """
        # 1-octet
        # +-----------+
        # | Version   |
        # +-----------+-----------+
        # | My Autonomous System  |
        # +-----------+-----------+
        # |       Hold Time       |
        # +-----------+-----------+-----------+-----------+
        # |                BGP Identifier                 |
        # +-----------+-----------+-----------+-----------+
        # |OptParm Len|
        # +-----------+-----------+-----------+-----------+
        # | Optional Parameters (variable)                |
        # +-----------+-----------+-----------+-----------+

        self.version = version
        self.asn = asn
        self.hold_time = hold_time
        self.bgp_id = bgp_id
        self.opt_para_len = opt_para_len
        self.opt_paras = opt_paras

        self.capa_dict = {}
        # used to store Capabilities {code: value}

    def parse(self, message):

        """Parses a BGP Open message"""

        try:
            self.version, self.asn, self.hold_time, \
                self.bgp_id, self.opt_para_len = struct.unpack('!BHHIB', message[:10])

        except:
            raise excp.MessageHeaderError(
                sub_error=bgp_cons.ERR_MSG_HDR_BAD_MSG_LEN,
                data=message[:10])

        self.bgp_id = str(netaddr.IPAddress(self.bgp_id))

        if self.version != 4:
            # Here we just support BGP-4
            raise excp.OpenMessageError(
                sub_error=bgp_cons.ERR_MSG_OPEN_UNSUP_VERSION,
                data=self.version)

        if isinstance(self.asn, float):
            self.asn = str(self.asn).split('.')
            self.asn = 65536 * (int(self.asn[0])) + int(self.asn[1])

        if self.asn in (0, 2 ** 16 - 1):
            # bad peer asn
            raise excp.OpenMessageError(
                sub_error=bgp_cons.ERR_MSG_OPEN_BAD_PEER_AS,
                data=self.asn)
        # Hold Time negotiation is out of this scope

        if self.bgp_id in (0, 2 ** 32 - 1):
            raise excp.OpenMessageError(
                sub_error=bgp_cons.ERR_MSG_OPEN_BAD_BGP_ID,
                data=self.bgp_id)
        # Optional Parameters
        if self.opt_para_len:

            self.opt_paras = message[10:]

            # While Loop: Parse one Optional Parameter(Capability) each time
            while self.opt_paras:

                # 1 octet     1 octet      variable
                # --------------------------------------+
                # para_type | para_length | para_value |
                # --------------------------------------+
                opt_para_type, opt_para_length = struct.unpack('!BB', self.opt_paras[:2])

                # Parameter Type 1: Authentication (deprecated) [RFC4271] [RFC5492]
                # Parameter Type 2: Capabilities [RFC5492]
                # Here we only support Type 2
                if opt_para_type != 2:
                    # if type is not type 2, return an suberror used to Notification
                    raise excp.OpenMessageError(
                        sub_error=bgp_cons.ERR_MSG_OPEN_UNSUP_OPT_PARAM,
                        data=message[10:])

                # ----------------------  Parse Capabilities ------------------#
                # capabilities belongs to one Optional Parameter Capability
                capabilities = self.opt_paras[2:opt_para_length + 2]

                while capabilities:

                    # ---- Parse every capability in this Optional Parameter
                    capability = Capability()
                    capability.parse(capabilities)

                    # (1) for 4 bytes ASN
                    if capability.capa_code == capability.FOUR_BYTES_ASN:
                        asn = struct.unpack('!I', capability.capa_value)[0]
                        self.asn = asn
                        self.capa_dict['four_bytes_as'] = True

                    # (2) Multiprotocol Extensions for BGP-4
                    elif capability.capa_code == capability.MULTIPROTOCOL_EXTENSIONS:
                        if 'afi_safi' not in self.capa_dict:
                            self.capa_dict['afi_safi'] = []
                        afi, res, safi = struct.unpack('!HBB', capability.capa_value)
                        self.capa_dict['afi_safi'].append((afi, safi))

                    # (3) Route Refresh
                    elif capability.capa_code == capability.ROUTE_REFRESH:
                        self.capa_dict['route_refresh'] = True

                    # (4) Cisco Route Refresh
                    elif capability.capa_code == capability.CISCO_ROUTE_REFRESH:
                        self.capa_dict['cisco_route_refresh'] = True

                    # (5) Graceful Restart
                    elif capability.capa_code == capability.GRACEFUL_RESTART:
                        self.capa_dict['graceful_restart'] = True

                    # (6) Cisco MultiSession
                    elif capability.capa_code == capability.CISCO_MULTISESSION_BGP:
                        self.capa_dict['cisco_multi_session'] = True

                    # (7) enhanced route refresh
                    elif capability.capa_code == capability.ENHANCED_ROUTE_REFRESH:
                        self.capa_dict['enhanced_route_refresh'] = True
                    # (8) add path
                    elif capability.capa_code == capability.ADD_PATH:
                        self.capa_dict['add_path'] = True
                    else:
                        self.capa_dict[str(capability.capa_code)] = capability.capa_value

                    capabilities = capabilities[2 + capability.capa_length:]

                # Go to next Optional Parameter
                self.opt_paras = self.opt_paras[opt_para_length + 2:]

            return {
                'Version': self.version,
                'ASN': self.asn,
                'holdTime': self.hold_time,
                'bgpID': self.bgp_id,
                'Capabilities': self.capa_dict
            }

    @staticmethod
    def construct_header(msg):

        """Prepends the mandatory header to a constructed BGP message
        # 16-octet     2-octet  1-octet
        #---------------+--------+---------+------+
        #    Maker      | Length |  Type   |  msg |
        #---------------+--------+---------+------+
        """
        return b'\xff'*16 + struct.pack('!HB',
                                        len(msg) + 19,
                                        1) + msg

    def construct(self, my_capability):

        """ Construct a BGP Open message """
        capas = b''
        # Construct Capabilities Optional Parameter (Parameter Type 2)
        if 'afi_safi' in my_capability:
            # Multiprotocol extentions capability
            capas += Capability(capa_code=1, capa_length=4).construct(my_capability)

        if my_capability.get('cisco_route_refresh'):
            # Cisco Route refresh capability
            capas += Capability(capa_code=128, capa_length=0).construct(my_capability)
        if my_capability.get('route_refresh'):
            # Route Refresh capability
            capas += Capability(capa_code=2, capa_length=0).construct(my_capability)

        # 4 bytes ASN
        if self.asn > 65535:
            capas += Capability(capa_code=65, capa_length=4, capa_value=self.asn).construct(my_capability)
            self.asn = 23456
        else:
            if my_capability.get('four_bytes_as'):
                capas += Capability(capa_code=65, capa_length=4, capa_value=self.asn).construct(my_capability)
        # for add path
        if my_capability.get('add_path'):
            capas += Capability(capa_code=69, capa_length=4, capa_value=my_capability['add_path']).construct()

        if my_capability.get('enhanced_route_refresh'):
            capas += Capability(capa_code=70, capa_length=0).construct()

        open_header = struct.pack('!BHHIB', self.version, self.asn, self.hold_time,
                                  self.bgp_id, len(capas))
        message = open_header + capas
        return self.construct_header(message)


# ========================================================================== Optional Parameters

class Capability(object):
    """
       The parameter contains one or more triples <Capability Code,
       Capability Length, Capability Value>, where each triple is encoded as
       shown below:

              +------------------------------+
              | Capability Code (1 octet)    |
              +------------------------------+
              | Capability Length (1 octet)  |
              +------------------------------+
              | Capability Value (variable)  |
              ~                              ~
              +------------------------------+

       The use and meaning of these fields are as follows:

          Capability Code:

             Capability Code is a one-octet unsigned binary integer that
             unambiguously identifies individual capabilities.

          Capability Length:

             Capability Length is a one-octet unsigned binary integer that
             contains the length of the Capability Value field in octets.

          Capability Value:

             Capability Value is a variable-length field that is interpreted
             according to the value of the Capability Code field.
    """

    # Capability Codes (IANA)
    # ---------------------------------------------------------------+
    #   Range  |   Registration Procedures  |          Notes        |
    #   1-63   |       IETF Review          |                       |
    #  64-127  |  First Come First Served   |                       |
    #  128-255 |   Reserved for Private Use |  IANA does not assign |
    # ---------------------------------------------------------------+

    # ---------------------------------------------------------------------------------------------------------------+
    #    Value  |                       Description                      |              Reference                   |
    #      0    |                        Reserved                        |              [RFC5492]                   |
    #      1    |           Multiprotocol Extensions for BGP-4           |              [RFC2858]                   |
    #      2    |           Route Refresh Capability for BGP-4           |              [RFC2918]                   |
    #      3    |           Outbound Route Filtering Capability          |              [RFC5291]                   |
    #      4    |       Multiple routes to a destination capability      |              [RFC3107]                   |
    #      5    |                 Extended Next Hop Encoding             |              [RFC5549]                   |
    #     6-63  |                        Unassigned                      |                                          |
    #      64   |                Graceful Restart Capability             |              [RFC4724]                   |
    #      65   |         Support for 4-octet AS number capability       |              [RFC4893]                   |
    #      66   |                   Deprecated (2003-03-06)              |                                          |
    #      67   |   Support for Dynamic Capability (capability specific) |     [draft-ietf-idr-dynamic-cap]         |
    #      68   |                 Multisession BGP Capability            |          [Chandra_Appanna]               |
    #      69   |                    ADD-PATH Capability                 |       [draft-ietf-idr-add-paths]         |
    #      70   |             Enhanced Route Refresh Capability          | [draft-keyur-bgp-enhanced-route-refresh] |
    #   71-127  |                       Unassigned                       |                                          |
    #   128-255 |                 Reserved for Private Use               |              [RFC5492]                   |
    # ---------------------------------------------------------------------------------------------------------------+

    # =================================================================== Capabilities
    # http://www.iana.org/assignments/capability-codes/

    RESERVED = 0x00  # [RFC5492]
    MULTIPROTOCOL_EXTENSIONS = 0x01  # [RFC2858]
    ROUTE_REFRESH = 0x02  # [RFC2918]
    OUTBOUND_ROUTE_FILTERING = 0x03  # [RFC5291]
    MULTIPLE_ROUTES = 0x04  # [RFC3107]
    EXTENDED_NEXT_HOP = 0x05  # [RFC5549]
    # 6-63      Unassigned
    GRACEFUL_RESTART = 0x40  # [RFC4724]
    FOUR_BYTES_ASN = 0x41  # [RFC4893]
    # 66 Deprecated
    DYNAMIC_CAPABILITY = 0x43  # [Chen]
    MULTISESSION_BGP = 0x44  # [Appanna]
    ADD_PATH = 0x45  # [draft-ietf-idr-add-paths]
    ENHANCED_ROUTE_REFRESH = 0x46
    # 70-127    Unassigned
    CISCO_ROUTE_REFRESH = 0x80  # I Can only find reference to this in the router logs
    # 128-255   Reserved for Private Use [RFC5492]

    CISCO_MULTISESSION_BGP = 0x83  # [Multisession BGP draft-ietf-idr-bgp-multisession-06]

    unassigned = range(70, 128)
    reserved = range(128, 256)

    def __init__(self, capa_code=None, capa_length=None, capa_value=None):

        """
          +------------------------------+
          | Capability Code (1 octet)    |
          +------------------------------+
          | Capability Length (1 octet)  |
          +------------------------------+
          | Capability Value (variable)  |
          ~                              ~
          +------------------------------+
        """
        self.capa_code = capa_code
        self.capa_length = capa_length
        self.capa_value = capa_value

    def parse(self, message):

        """
        Partition Capabilities message one by one
        """
        try:
            self.capa_code, self.capa_length = struct.unpack('!BB', message[:2])
        except:
            raise excp.OpenMessageError(
                sub_error=bgp_cons.ERR_MSG_HDR_BAD_MSG_LEN,
                data=message[:2])
        self.capa_value = message[2:self.capa_length + 2]

    def construct(self, my_capability=None):

        """ Construct a capability PDU """

        # for 4 bytes as
        if self.capa_code == self.FOUR_BYTES_ASN:
            return struct.pack('!BBBBI', 2, 6, self.FOUR_BYTES_ASN, self.capa_length, self.capa_value)

        # for route refresh
        elif self.capa_code == self.ROUTE_REFRESH:
            return struct.pack('!BBBB', 2, 2, self.ROUTE_REFRESH, 0)

        # for cisco route refresh
        elif self.capa_code == self.CISCO_ROUTE_REFRESH:
            return struct.pack('!BBBB', 2, 2, self.CISCO_ROUTE_REFRESH, 0)

        # graceful restart
        elif self.capa_code == self.GRACEFUL_RESTART:
            return struct.pack('!BBBB', 2, 2, self.GRACEFUL_RESTART, 0)

        # for multiprotocol extentions
        elif self.capa_code == self.MULTIPROTOCOL_EXTENSIONS:
            # <ipv4,unicast> and <ipv4,mplsvpn>
            afisafi = b''
            for (afi, safi) in my_capability['afi_safi']:
                afisafi += struct.pack('!BBBBHBB', 2, 6, self.MULTIPROTOCOL_EXTENSIONS, 4, afi, 0, safi)
            return afisafi
        # for add path
        elif self.capa_code == self.ADD_PATH:
            add_path = struct.pack('!BBBBHBB', 2, 6, self.ADD_PATH, self.capa_length, 1, 1, self.capa_value)
            return add_path

        elif self.capa_code == self.ENHANCED_ROUTE_REFRESH:
            return struct.pack('!BBBB', 2, 2, self.ENHANCED_ROUTE_REFRESH, 0)



# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Twisted message queue factory
reference from https://github.com/pika/pika/blob/master/examples/twisted_service.py
"""

import logging

import pika
from twisted.internet import protocol
from twisted.internet import reactor

from yabgp.channel.protocol import PikaProtocol

LOG = logging.getLogger(__name__)


class PikaFactory(protocol.ReconnectingClientFactory):

    def __init__(self, host='localhost', port=5672, userid='guest', password='guest'):
        self.parameters = pika.ConnectionParameters(
            host=host, port=port,
            credentials=pika.PlainCredentials(userid, password)
        )
        self.client = None
        self.queued_messages = []
        self.read_list = []

    def startedConnecting(self, connector):
        LOG.info('Started to connect to AMQP')

    def buildProtocol(self, addr):
        self.resetDelay()
        LOG.info('Connected AMQP')
        self.client = PikaProtocol(self.parameters)
        self.client.factory = self
        self.client.ready.addCallback(self.client.connected)
        return self.client

    def clientConnectionLost(self, connector, reason):
        LOG.info('Lost connection.  Reason: %s', reason)
        protocol.ReconnectingClientFactory.clientConnectionLost(self, connector, reason.getErrorMessage())

    def clientConnectionFailed(self, connector, reason):
        LOG.info('Connection failed. Reason: %s', reason)
        protocol.ReconnectingClientFactory.clientConnectionFailed(self, connector, reason.getErrorMessage())

    def send_message(self, exchange=None, routing_key=None, message=None):
        self.queued_messages.append((exchange, routing_key, message))
        if self.client is not None:
            self.client.send()

    def read_messages(self, exchange, routing_key, callback):
        """Configure an exchange to be read from."""
        self.read_list.append((exchange, routing_key, callback))
        if self.client is not None:
            self.client.read(exchange, routing_key, callback)

    def connect(self):

        try:
            reactor.connectTCP(
                host=self.parameters.host,
                port=self.parameters.port,
                factory=self)
        except Exception as e:
            LOG.error(e)



# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Blueprint for version 1 of API
"""
import logging

from flask.ext.httpauth import HTTPBasicAuth
from flask import Blueprint
import flask
from oslo.config import cfg

from yabgp.api import utils as api_utils

LOG = logging.getLogger(__name__)
blueprint = Blueprint('v1', __name__)
auth = HTTPBasicAuth()


@auth.get_password
def get_pw(username):
    if username == cfg.CONF.rest.username:
        return cfg.CONF.rest.password
    return None


@blueprint.route('/')
def root():
    """
    v1 api root. Get the api status.

    **Example request**:

    .. sourcecode:: http

      GET /v1 HTTP/1.1
      Host: example.com
      Accept: application/json

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Vary: Accept
      Content-Type: text/json
      {
        "status": "stable",
        "updated": "2015-01-22T00:00:00Z",
        "version": "v1"
      }

    :status 200: the api can work.
    """
    intro = {
        "status": "stable",
        "updated": "2015-01-22T00:00:00Z",
        "version": "v1"}
    return flask.jsonify(intro)


@blueprint.route('/peers', methods=['GET'])
@auth.login_required
def peers():
    """
    Get all peers realtime running information, include basic configurations and fsm state.

    **Example request**

    .. sourcecode:: http

      GET /v1/peers HTTP/1.1
      Host: example.com
      Accept: application/json

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Vary: Accept
      Content-Type: text/json

      {
          "peers": [
              {
                  "fsm": "ESTABLISHED",
                  "local_addr": "100.100.0.1",
                  "local_as": 65022,
                  "remote_addr": "100.100.9.1",
                  "remote_as": 65022,
                  "uptime": 106810.47324299812
              },
              {
                  "fsm": "ESTABLISHED",
                  "local_addr": "100.100.0.1",
                  "local_as": 65022,
                  "remote_addr": "100.100.9.1",
                  "remote_as": 65022,
                  "uptime": 106810.47324299812
              }
          ]
      }

    :status 200: the api can work.
    """
    return flask.jsonify(api_utils.get_peer_conf_and_state())


@blueprint.route('/peer/<peer_ip>/state')
@auth.login_required
def peer(peer_ip):
    """
    Get one peer's running information, include basic configurations and fsm state.

    **Example request**

    .. sourcecode:: http

      GET /v1/peer/10.124.1.245/state HTTP/1.1
      Host: example.com
      Accept: application/json

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Vary: Accept
      Content-Type: text/json
      {
        "peer": {
            "fsm": "ESTABLISHED",
            "local_addr": "10.75.44.11",
            "local_as": 23650,
            "remote_addr": "10.124.1.245",
            "remote_as": 23650,
            "uptime": 7.913731813430786
            }
        }

    :param peer_ip: peer ip address
    :status 200: the api can work.

    """
    return flask.jsonify(api_utils.get_peer_conf_and_state(peer_ip))


@blueprint.route('/peer/<peer_ip>/statistic')
@auth.login_required
def get_peer_statistic(peer_ip):
    """
    Get one peer's message statistic, include sending and receiving.

    **Example request**

    .. sourcecode:: http

      GET /v1/peer/10.124.1.245/statistic HTTP/1.1
      Host: example.com
      Accept: application/json

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Vary: Accept
      Content-Type: text/json
      {
          "receive": {
              "Keepalives": 3,
              "Notifications": 0,
              "Opens": 1,
              "Route Refresh": 0,
              "Updates": 5
          },
          "send": {
              "Keepalives": 3,
              "Notifications": 0,
              "Opens": 1,
              "Route Refresh": 0,
              "Updates": 0
          }
      }

    :param peer_ip: peer ip address
    :status 200: the api can work.

    """
    return flask.jsonify(api_utils.get_peer_msg_statistic(peer_ip))


@blueprint.route('/peer/<peer_ip>/send/route-refresh', methods=['POST'])
@auth.login_required
def send_route_refresh(peer_ip):
    """
    Try to send BGP Route Refresh message to a peer

    **Example request**

    .. sourcecode:: http

      POST /v1/peer/10.124.1.245/send/route-refresh HTTP/1.1
      Host: example.com
      Accept: application/json
      POST Data:
      {
        "afi": 1,
        "safi": 1,
        "res": 0
      }
      `res` is optional, the default value is 0.

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Vary: Accept
      Content-Type: text/json
      {
        "status": True
      }

    Success f the status is `True` in the reponse, otherwise, the status is `False`. If the status is
    Flase, there will be a code tell you why. like:

    .. sourcecode:: http

        {
            "status": False,
            "code": "please check your post data"
        }

    :param peer_ip: peer ip address
    :status 200: the api can work.
    """
    LOG.debug('Try to send route refresh to peer %s', peer_ip)
    json_request = flask.request.get_json()
    if 'afi' in json_request and 'safi' in json_request:
        if 'res' not in json_request:
            res = 0
        else:
            res = json_request['res']
        result = api_utils.send_route_refresh(
            peer_ip=peer_ip, afi=json_request['afi'], safi=json_request['safi'], res=res)
        return flask.jsonify(result)
    return flask.jsonify({
        'status': False,
        'code': 'please check your post data'
    })


@blueprint.route('/peer/<peer_ip>/send/update', methods=['POST'])
@auth.login_required
def send_update_message(peer_ip):
    """
    Try to send BGP update message to the peer. Both update nlri and withdraw nlri treated as Update.

    **Example request**

    .. sourcecode:: http

      POST /v1/peer/10.124.1.245/send/update HTTP/1.1
      Host: example.com
      Accept: application/json

    Post data example for IPv4 update and withdraw

    .. sourcecode:: http

        # update
        {
            "attr":{
                "1": 0,
                "2": [],
                "3": "192.0.2.1",
                "5": 100,
                "8": ["NO_EXPORT"]
        },
            "nlri": ["172.20.1.0/24", "172.20.2.0/24"]
        }
        # withdraw
        {
            "withdraw": ["172.20.1.0/24", "172.20.2.0/24"]
        }


    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Vary: Accept
      Content-Type: text/json
      {
        "status": True
      }

    Success f the status is `True` in the reponse, otherwise, the status is `False`. If the status is
    Flase, there will be a code tell you why. like:

    .. sourcecode:: http

        {
            "status": False,
            "code": "please check your post data"
        }

    :param peer_ip: peer ip address
    :status 200: the api can work.

    """
    LOG.debug('Try to send update message to peer %s', peer_ip)
    json_request = flask.request.get_json()
    attr = json_request.get('attr')
    nlri = json_request.get('nlri')
    withdraw = json_request.get('withdraw')
    if attr:
        attr = {int(k): v for k, v in attr.items()}
    if (attr and nlri) or withdraw:
        return flask.jsonify(api_utils.send_update(peer_ip, attr, nlri, withdraw))
    else:
        return flask.jsonify({
            'status': False,
            'code': 'please check your post data'
        })



# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

""" Test Route Refresh message"""

import unittest
from yabgp.message.route_refresh import RouteRefresh
from yabgp.common.constants import MSG_CISCOROUTEREFRESH
from yabgp.common.constants import MSG_ROUTEREFRESH


class TestRouteRefresh(unittest.TestCase):

    def test_parse(self):

        msg_hex = b'\x00\x01\x00\x80'
        msg_parsed = RouteRefresh().parse(msg_hex)
        msg_hoped = (1, 0, 128)
        self.assertEqual(msg_hoped, msg_parsed)

    def test_construct_rfc_route_refresh(self):

        msg = (1, 0, 128)
        msg_hex = RouteRefresh(afi=msg[0], res=msg[1], safi=msg[2]).construct(MSG_ROUTEREFRESH)
        msg_hoped = b'\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x00\x17\x05\x00\x01\x00\x80'
        self.assertEqual(msg_hoped, msg_hex)

    def test_construct_cisco_route_refresh(self):

        msg = (1, 0, 128)
        msg_hex = RouteRefresh(afi=msg[0], res=msg[1], safi=msg[2]).construct(MSG_CISCOROUTEREFRESH)
        msg_hoped = b'\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x00\x17\x80\x00\x01\x00\x80'
        self.assertEqual(msg_hoped, msg_hex)

if __name__ == '__main__':
    unittest.main()



# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import struct

from yabgp.message.attribute import Attribute
from yabgp.message.attribute import AttributeID
from yabgp.message.attribute import AttributeFlag
from yabgp.common import constants as bgp_cons
from yabgp.common import exception as excep


class MED(Attribute):
    """
    This is an optional non-transitive attribute that is a
    four-octet unsigned integer. The value of this attribute
    MAY be used by a BGP speaker's Decision Process to
    discriminate among multiple entry points to a neighboring
    autonomous system.
    """

    ID = AttributeID.MULTI_EXIT_DISC
    FLAG = AttributeFlag.OPTIONAL
    MULTIPLE = False

    @classmethod
    def parse(cls, value):
        """
        parse BGP med attributes
        :param value: raw binary value
        """
        try:
            return struct.unpack('!I', value)[0]
        except:
            raise excep.UpdateMessageError(
                sub_error=bgp_cons.ERR_MSG_UPDATE_ATTR_LEN,
                data=value)

    @classmethod
    def construct(cls, value):
        """
        encode BGP med attributes
        :param value:
        """
        try:
            return struct.pack('!B', cls.FLAG) + struct.pack('!B', cls.ID) \
                + struct.pack('!B', 4) + struct.pack('!I', value)
        except Exception:
            raise excep.UpdateMessageError(
                sub_error=bgp_cons.ERR_MSG_UPDATE_ATTR_LEN,
                data='')



# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import struct

import netaddr

from yabgp.message.attribute import Attribute
from yabgp.message.attribute import AttributeID
from yabgp.message.attribute import AttributeFlag
from yabgp.common import constants as bgp_cons
from yabgp.common import exception as excep


class Aggregator(Attribute):
    """
        AGGREGATOR is an optional transitive attribute of length 6.
    The attribute contains the last AS number that formed the
    aggregate route (encoded as 2 octets), followed by the IP
    address of the BGP speaker that formed the aggregate route
    (encoded as 4 octets). This SHOULD be the same address as
    the one used for the BGP Identifier of the speaker
    """

    ID = AttributeID.AGGREGATOR
    FLAG = AttributeFlag.OPTIONAL + AttributeFlag.TRANSITIVE

    @classmethod
    def parse(cls, value, asn4=False):

        """
        Parse Aggregator attributes.
        :param value: raw binary value
        :param asn4: support 4 bytes asn or not
        """

        if not asn4:
            try:
                asn = struct.unpack('!H', value[:2])[0]
                aggregator = str(netaddr.IPAddress(struct.unpack('!I', value[2:])[0]))
            except Exception:
                raise excep.UpdateMessageError(
                    sub_error=bgp_cons.ERR_MSG_UPDATE_ATTR_LEN,
                    data=value)
        else:
            # 4 bytes ASN
            try:
                asn = struct.unpack('!I', value[:4])[0]
                aggregator = str(netaddr.IPAddress(struct.unpack('!I', value[4:])[0]))
            except Exception:
                raise excep.UpdateMessageError(
                    sub_error=bgp_cons.ERR_MSG_UPDATE_ATTR_LEN,
                    data=value)

        return asn, aggregator

    @classmethod
    def construct(cls, value, asn4=False):

        """
        Construct aggregator.
        :param value:
        :param asn4:
        """
        try:
            if asn4:
                agg_raw = struct.pack('!I', value[0]) + netaddr.IPAddress(value[1]).packed
            else:
                agg_raw = struct.pack('!H', value[0]) + netaddr.IPAddress(value[1]).packed

            return struct.pack('!B', cls.FLAG) + struct.pack('!B', cls.ID) \
                + struct.pack('!B', len(agg_raw)) + agg_raw
        except Exception:
            raise excep.UpdateMessageError(
                sub_error=bgp_cons.ERR_MSG_UPDATE_ATTR_LEN,
                data=value)



# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

""" Test BGP KeepAlive message"""

import unittest

from yabgp.message.keepalive import KeepAlive
from yabgp.common.exception import MessageHeaderError


class TestKeepAlive(unittest.TestCase):

    def test_parse(self):
        msg_hex = ''
        keepalive_msg = KeepAlive().parse(msg_hex)
        self.assertEqual(None, keepalive_msg)

    def test_parse_msg_header_error(self):
        self.assertRaises(MessageHeaderError, KeepAlive().parse, b'\x00')

    def test_construct(self):

        msg_hex = KeepAlive().construct()
        hope_msg = b'\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x00\x13\x04'
        self.assertEqual(hope_msg, msg_hex)

if __name__ == '__main__':
    unittest.main()



# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

""" Test extended Community attribute """

import unittest

from yabgp.common import constants as bgp_cons
from yabgp.common import exception as excep
from yabgp.message.attribute.extcommunity import ExtCommunity


class TestExtCommunity(unittest.TestCase):

    def test_parse_rt0(self):
        # Route Target,Format AS(2bytes):AN(4bytes)
        ext_community = ExtCommunity.parse(value=b'\x00\x02\x00\x64\x00\x00\x00\x0c')
        self.assertEqual([(bgp_cons.BGP_EXT_COM_RT_0, '100:12')], ext_community)

    def test_construct_rt0(self):
        ext_community = ExtCommunity.construct(value=[(bgp_cons.BGP_EXT_COM_RT_0, '100:12')])
        self.assertEqual(b'\xc0\x10\x08\x00\x02\x00\x64\x00\x00\x00\x0c', ext_community)

    def test_parse_rt1(self):
        # Route Target,Format IPv4 address(4bytes):AN(2bytes)
        ext_community = ExtCommunity.parse(value=b'\x01\x02\x0a\x0a\x0a\x0a\x00\x0c')
        self.assertEqual([(bgp_cons.BGP_EXT_COM_RT_1, '10.10.10.10:12')], ext_community)

    def test_construct_rt1(self):
        ext_community = ExtCommunity.construct(value=[(bgp_cons.BGP_EXT_COM_RT_1, '10.10.10.10:12')])
        self.assertEqual(b'\xc0\x10\x08\x01\x02\x0a\x0a\x0a\x0a\x00\x0c', ext_community)

    def test_parse_rt2(self):
        # Route Target,Format AS(4bytes):AN(2bytes)
        ext_community = ExtCommunity.parse(value=b'\x02\x02\x00\x01\x00\x01\x00\x0c')
        self.assertEqual([(bgp_cons.BGP_EXT_COM_RT_2, '65537:12')], ext_community)

    def test_construct_rt2(self):
        ext_community = ExtCommunity.construct(value=[(bgp_cons.BGP_EXT_COM_RT_2, '65537:12')])
        self.assertEqual(b'\xc0\x10\x08\x02\x02\x00\x01\x00\x01\x00\x0c', ext_community)

    def test_parse_ro0(self):
        # Route Origin,Format AS(2bytes):AN(4bytes)
        ext_community = ExtCommunity.parse(value=b'\x00\x03\x00\x64\x00\x00\x00\x0c')
        self.assertEqual([(bgp_cons.BGP_EXT_COM_RO_0, '100:12')], ext_community)

    def test_construct_ro0(self):
        ext_community = ExtCommunity.construct(value=[(bgp_cons.BGP_EXT_COM_RO_0, '100:12')])
        self.assertEqual(b'\xc0\x10\x08\x00\x03\x00\x64\x00\x00\x00\x0c', ext_community)

    def test_parse_ro1(self):
        # Route Origin,Format IPv4 address(4bytes):AN(2bytes)
        ext_community = ExtCommunity.parse(value=b'\x01\x03\x0a\x0a\x0a\x0a\x00\x0c')
        self.assertEqual([(bgp_cons.BGP_EXT_COM_RO_1, '10.10.10.10:12')], ext_community)

    def test_construct_ro1(self):
        ext_community = ExtCommunity.construct(value=[(bgp_cons.BGP_EXT_COM_RO_1, '10.10.10.10:12')])
        self.assertEqual(b'\xc0\x10\x08\x01\x03\x0a\x0a\x0a\x0a\x00\x0c', ext_community)

    def test_parse_ro2(self):
        # Route Origin,Format AS(4bytes):AN(2bytes)
        ext_community = ExtCommunity.parse(value=b'\x02\x03\x00\x01\x00\x01\x00\x0c')
        self.assertEqual([(bgp_cons.BGP_EXT_COM_RO_2, '65537:12')], ext_community)

    def test_construct_ro2(self):
        ext_community = ExtCommunity.construct(value=[(bgp_cons.BGP_EXT_COM_RO_2, '65537:12')])
        self.assertEqual(b'\xc0\x10\x08\x02\x03\x00\x01\x00\x01\x00\x0c', ext_community)

    def test_parse_invalid_length(self):
        # invalid length
        self.assertRaises(excep.UpdateMessageError, ExtCommunity.parse,
                          b'\x00\x00\x02\x00\x64\x00\x00\x00\x0c')
        try:
            ExtCommunity.parse(value=b'\x00\x00\x02\x00\x64\x00\x00\x00\x0c')
        except excep.UpdateMessageError as e:
            self.assertEqual(bgp_cons.ERR_MSG_UPDATE_ATTR_LEN, e.sub_error)

    def test_parse_unknow(self):
        # unknow
        hex_tmp = b'\x09\x03\x00\x01\x00\x01\x00\x0c'
        ext_community = ExtCommunity.parse(value=hex_tmp)
        self.assertEqual(bgp_cons.BGP_EXT_COM_UNKNOW, ext_community[0][0])

if __name__ == '__main__':
    unittest.main()



# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

""" BGP KeepAlive message"""

import struct

from yabgp.common.exception import MessageHeaderError
from yabgp.common.constants import ERR_MSG_HDR_BAD_MSG_LEN


class KeepAlive(object):

    """
    KEEPALIVE messages are exchanged between peers often
    enough not to cause the Hold Timer to expire
    """
    MSG_KEEPALIVE = 4

    @staticmethod
    def parse(msg):
        """
        Parse keepalive message

        :param msg: input raw binary message data
        """
        if len(msg) != 0:
            raise MessageHeaderError(
                sub_error=ERR_MSG_HDR_BAD_MSG_LEN,
                data='')

    @staticmethod
    def construct_header():

        """Prepends the mandatory header to a constructed BGP message
        """
        #    16-octet     2-octet  1-octet
        # ---------------+--------+---------+------+
        #    Maker      | Length |  Type   |  msg |
        # ---------------+--------+---------+------+
        return b'\xff'*16 + struct.pack('!HB', 19, 4)

    def construct(self):
        """
        Construct a keepalive message
        """
        return self.construct_header()



# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import struct
import binascii

import netaddr

from yabgp.message.attribute import Attribute
from yabgp.message.attribute import AttributeID
from yabgp.message.attribute import AttributeFlag
from yabgp.common import constants as bgp_cons
from yabgp.common import exception as excep


class NextHop(Attribute):
    """
        This is a well-known mandatory attribute that defines the
    (unicast) IP address of the router that SHOULD be used as
    the next hop to the destinations listed in the Network Layer
    Reachability Information field of the UPDATE message.
    """

    ID = AttributeID.NEXT_HOP
    FLAG = AttributeFlag.TRANSITIVE
    MULTIPLE = False

    @classmethod
    def parse(cls, value):

        """
        Parse BGP nexthop.
        :param value: raw binary value
        """
        if len(value) % 4 == 0:
            next_hop = str(netaddr.IPAddress(int(binascii.b2a_hex(value[0:4]), 16)))
            return next_hop
        else:
            # Error process
            raise excep.UpdateMessageError(
                sub_error=bgp_cons.ERR_MSG_UPDATE_ATTR_LEN,
                data=value)

    @classmethod
    def construct(cls, value):
        """
        encode BGP nexthop attribute.
        :param value: ipv4 format string like 1.1.1.1
        """
        try:
            if netaddr.IPAddress(value).version == 4:
                ip_addr_raw = netaddr.IPAddress(value).packed
                return struct.pack('!B', cls.FLAG) + struct.pack('!B', cls.ID) \
                    + struct.pack('!B', len(ip_addr_raw)) + ip_addr_raw
            else:
                raise excep.UpdateMessageError(
                    sub_error=bgp_cons.ERR_MSG_UPDATE_INVALID_NEXTHOP,
                    data=value)
        except Exception:
                raise excep.UpdateMessageError(
                    sub_error=bgp_cons.ERR_MSG_UPDATE_INVALID_NEXTHOP,
                    data=value)



# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

""" Test Notification message"""

import unittest


from yabgp.message.notification import Notification


class TestNotification(unittest.TestCase):

    def test_parse(self):
        msg_hex = b'\x03\x05\x00\x00'
        noti_msg = Notification().parse(msg_hex)
        self.assertEqual((3, 5, b'\x00\x00'), noti_msg)

    def test_construct(self):

        msg_hex = Notification().construct(error=3, suberror=5, data=b'\x00\x00')
        hope_msg = b'\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff' \
                   b'\xff\xff\x00\x17\x03\x03\x05\x00\x00'
        self.assertEqual(hope_msg, msg_hex)

if __name__ == '__main__':
    unittest.main()



# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

""" Test Atomic Aggregate attribute """

import unittest

from yabgp.common.exception import UpdateMessageError
from yabgp.common.constants import ERR_MSG_UPDATE_OPTIONAL_ATTR
from yabgp.message.attribute.atomicaggregate import AtomicAggregate


class TestAtomicAggregate(unittest.TestCase):

    def test_parse(self):

        atomicaggregate = AtomicAggregate.parse(value=b'')
        self.assertEqual(atomicaggregate, b'')

        # invalid value
        self.assertRaises(UpdateMessageError, AtomicAggregate.parse,
                          b'\x01')
        try:
            AtomicAggregate.parse(b'\x01')
        except UpdateMessageError as e:
            self.assertEqual(e.sub_error, ERR_MSG_UPDATE_OPTIONAL_ATTR)

    def test_construct(self):

        atomicaggregate = AtomicAggregate.construct(value=b'')
        self.assertEqual(b'\xc0\x06\x00', atomicaggregate)

    def test_construct_exception(self):
        self.assertRaises(UpdateMessageError, AtomicAggregate.construct, b'\x00')

if __name__ == '__main__':
    unittest.main()



# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

""" BGP Exception """

_FATAL_EXCEPTION_FORMAT_ERRORS = False


class BGPException(Exception):
    """Base BGP Exception.
    """
    message = "An unknown exception occurred."

    def __init__(self, **kwargs):
        try:
            super(BGPException, self).__init__(self.message % kwargs)
            self.msg = self.message % kwargs
        except Exception:
            if _FATAL_EXCEPTION_FORMAT_ERRORS:
                raise
            else:
                # at least get the core message out if something happened
                super(BGPException, self).__init__(self.message)

    def __unicode__(self):
        return unicode(self.msg)


class NotificationSent(Exception):
    """BGP Notification Exception.
    """
    message = "Unknow Notification exception occurred"
    error = 0

    def __init__(self, sub_error, data=''):
        try:
            super(NotificationSent, self).__init__()
            self.msg = self.message % {'sub_error': sub_error, 'data': data}
            self.sub_error = sub_error
            self.data = data
        except Exception:
            if _FATAL_EXCEPTION_FORMAT_ERRORS:
                raise
            else:
                # at least get the core message out if something happened
                super(NotificationSent, self).__init__(self.message)

    def __unicode__(self):
        return unicode(self.msg)


class MessageHeaderError(NotificationSent):
    error = 1
    message = "BGP Message Header Error, sub error:%(sub_error)s, data:%(data)s"


class OpenMessageError(NotificationSent):
    error = 2
    message = "BGP Open Message Error, sub error:%(sub_error)s, data:%(data)s"


class UpdateMessageError(NotificationSent):
    error = 3
    message = "BGP Update Message Error, sub error:%(sub_error)s, data:%(data)s"


class HoldTimerExpiredError(NotificationSent):
    error = 4
    message = "BGP Hold Timer Expired, sub error:%(sub_error)s, data:%(data)s"


class FSMError(NotificationSent):
    error = 5
    message = "BGP FSM Error, sub error:%(sub_error)s, data:%(data)s"


class ErrCease(NotificationSent):
    error = 6
    message = "BGP ERR CEASE Error, sub error:%(sub_error)s, data:%(data)s"



# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from struct import pack


class AttributeFlag(int):
    """
    +-------------------------------------------+
    | bit 0 | bit 1 | bit 2 | bit 3 | bit 4 - 7 |
    +-------------------------------------------+
    bit 0:
        It defines whether the attribute is optional (if set to 1)
    or well-known (if set to 0).
    bit 1:
        It defines whether an optional attribute is transitive
    (if set to 1) or non-transitive (if set to 0).
        For well-known attributes, the Transitive bit MUST be set to 1.
    bit 2:
        It defines whether the information contained in the optional
    transitive attribute is partial (if set to 1) or complete (if set to 0).
    For well-known attributes and for optional non-transitive attributes,
    the Partial bit MUST be set to 0.
    bit 3:
        It defines whether the Attribute Length is one octet
    (if set to 0) or two octets (if set to 1).
    bit 4 - 7
        The lower-order four bits of the Attribute Flags octet are
    unused. They MUST be zero when sent and MUST be ignored when
    received.
    """

    EXTENDED_LENGTH = 0x10  # 16  RFC 4271
    PARTIAL = 0x20  # 32  RFC 4271
    TRANSITIVE = 0x40  # 64  RFC 4271
    OPTIONAL = 0x80  # 128 RFC 4271

    def __str_(self):
        r = []
        v = int(self)
        if v & 0x10:
            r.append("EXTENDED_LENGTH")
            v -= 0x10
        if v & 0x20:
            r.append("PARTIAL")
            v -= 0x20
        if v & 0x40:
            r.append("TRANSITIVE")
            v -= 0x40
        if v & 0x80:
            r.append("OPTIONAL")
            v -= 0x80
        if v:
            r.append("UNKNOWN %s" % hex(v))
        return " ".join(r)


# ============================================ Attribute Type code
#  http://www.iana.org/assignments/bgp-parameters/bgp-parameters.xml
#  update 2011-9-26


class AttributeID(int):
    """
    Attribute Type Code
    """

    ORIGIN = 0x01  # 1   [RFC4271]
    AS_PATH = 0x02  # 2   [RFC4271]
    NEXT_HOP = 0x03  # 3   [RFC4271]
    MULTI_EXIT_DISC = 0x04  # 4   [RFC4271]
    LOCAL_PREF = 0x05  # 5   [RFC4271]
    ATOMIC_AGGREGATE = 0x06  # 6   [RFC4271]
    AGGREGATOR = 0x07  # 7   [RFC4271]
    COMMUNITY = 0x08  # 8   [RFC1997]
    ORIGINATOR_ID = 0x09  # 9   [RFC4456]
    CLUSTER_LIST = 0x0a  # 10  [RFC4456]
    # DPA                                     = 0x0b  # 11  [Chen, E., Bates, T., "Destination Preference Attribute for
    # BGP", Work in progress, March 1996.]
    # ADVERTISER (Historic)                   = 0x0c  # 12  [RFC1863][RFC4223]
    # RCID_PATH / CLUSTER_ID (Historic)       = 0x0d  # 13  [RFC1863][RFC4223]
    MP_REACH_NLRI = 0x0e  # 14  [RFC4760]
    MP_UNREACH_NLRI = 0x0f  # 15  [RFC4760]
    EXTENDED_COMMUNITY = 0x10  # 16  [Eric_Rosen][draft-ramachandra-bgp-ext-communities-00][RFC4360]
    AS4_PATH = 0x11  # 17  [RFC4893]
    AS4_AGGREGATOR = 0x12  # 18  [RFC4893]
    # SAFI_Specific_Attribute                 = 0x13  # 19  [Gargi_Nalawade]
    # [draft-kapoor-nalawade-idr-bgp-ssa-00][draft-nalawade-idr-mdt-safi-00][draft-wijnands-mt-discovery-00]
    # (SSA)(deprecated)
    # Connector_Attribute                     = 0x14  # 20  [RFC6037] (deprecated)
    # AS_PATHLIMIT (deprecated)               = 0x15  # 21  [draft-ietf-idr-as-pathlimit]
    PMSI_TUNNEL = 0x16  # 22  [RFC-ietf-l3vpn-2547bis-mcast-bgp-08]
    Tunnel_Encapsulation_Attribute = 0x17  # 23  [RFC5512]
    Traffic_Engineering = 0x18  # 24  [RFC5543]
    IPv6_Address_Specific_Extended_Community = 0x19  # 25  [RFC5701]
    # AIGP                                    = 0x1a  # 26  [draft-ietf-idr-aigp](TEMPORARY - expires 2012-03-14)

    ATTR_SET = 0x80  # 128 [RFC6368]

    Unassigned = list(range(27, 127)) + list(range(129, 254))
    Reserved_For_Development = 255

    def __str__(self):

        # This should move within the classes and not be here
        if self == 0x01:
            return "ORIGIN"
        if self == 0x02:
            return "AS_PATH"
        if self == 0x03:
            return "NEXT_HOP"
        if self == 0x04:
            return "MULTI_EXIT_DISC"
        if self == 0x05:
            return "LOCAL_PREFERENCE"
        if self == 0x06:
            return "ATOMIC_AGGREGATE"
        if self == 0x07:
            return "AGGREGATOR"
        if self == 0x08:
            return "COMMUNITY"
        if self == 0x09:
            return "ORIGINATOR_ID"
        if self == 0x0a:
            return "CLUSTER_LIST"
        if self == 0x10:
            return "EXTENDED_COMMUNITY"
        if self == 0x11:
            return "AS4_PATH"
        if self == 0x12:
            return "AS4_AGGREGATOR"
        if self == 0x0e:
            return "MP_REACH_NLRI"
        if self == 0x0f:
            return "MP_UNREACH_NLRI"
        if self == 0xffff:
            return "INTERNAL SPLIT"
        return 'UNKNOWN ATTRIBUTE (%s)' % hex(self)


class Attribute(object):
    """
    Base class for all BGP attribute classes
    Attribute instances are (meant to be) immutable once initialized
    """
    ID = 0x00
    FLAG = 0x00

    def _attribute(self, value):
        flag = self.FLAG
        if flag & AttributeFlag.OPTIONAL and not value:
            return ''
        length = len(value)
        if length > 0xFF:
            flag &= AttributeFlag.EXTENDED_LENGTH
        if flag & AttributeFlag.EXTENDED_LENGTH:
            len_value = pack('!H', length)[0]
        else:
            len_value = chr(length)
        return "%s%s%s%s" % (chr(flag), chr(self.ID), len_value, value)

    def __eq__(self, other):
        return self.ID == other.ID



# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""BGP extended community
"""

import struct
import logging

import netaddr

from yabgp.message.attribute import Attribute
from yabgp.message.attribute import AttributeID
from yabgp.message.attribute import AttributeFlag
from yabgp.common import exception as excep
from yabgp.common import constants as bgp_cons

LOG = logging.getLogger()


class ExtCommunity(Attribute):
    """
        The Extended Communities Attribute is a transitive optional BGP
    attribute, with the Type Code 16. The attribute consists of a set of
    "extended communities". All routes with the Extended Communities
    attribute belong to the communities listed in the attribute.
    (RFC 4360 Page2)
    http://www.iana.org/assignments/bgp-extended-communities
    """

    # Extended Community Type
    # IANA    0x0000 - 0x7fff
    # private 0x8000 - 0xffff

    ID = AttributeID.EXTENDED_COMMUNITY
    FLAG = AttributeFlag.OPTIONAL + AttributeFlag.TRANSITIVE

    @classmethod
    def parse(cls, value):

        """
        Each Extended Community is encoded as an 8-octet quantity, as
        follows:
        - Type Field : 1 or 2 octets
        - Value Field : Remaining octets
        0                   1                   2                   3
        0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        | Type high | Type low(*) |                                     |
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+ Value                         |
        |                                                               |
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        Parse Extended Community attributes.
        :param value : value
        """
        # devide every ext community
        if len(value) % 8 != 0:
            raise excep.UpdateMessageError(
                sub_error=bgp_cons.ERR_MSG_UPDATE_ATTR_LEN,
                data=value)

        ext_community = []
        while value:
            comm_type, subtype = struct.unpack('!BB', value[0:2])
            value_tmp = value[2:8]

            comm_code = comm_type * 256 + subtype

            # Route Target
            if comm_code == bgp_cons.BGP_EXT_COM_RT_0:
                # Route Target, Format AS(2bytes):AN(4bytes)
                asn, an = struct.unpack('!HI', value_tmp)
                ext_community.append((bgp_cons.BGP_EXT_COM_RT_0, '%s:%s' % (asn, an)))

            elif comm_code == bgp_cons.BGP_EXT_COM_RT_1:
                # Route Target,Format IPv4 address(4bytes):AN(2bytes)
                ipv4 = str(netaddr.IPAddress(struct.unpack('!I', value_tmp[0:4])[0]))
                an = struct.unpack('!H', value_tmp[4:])[0]
                ext_community.append((bgp_cons.BGP_EXT_COM_RT_1, '%s:%s' % (ipv4, an)))

            elif comm_code == bgp_cons.BGP_EXT_COM_RT_2:
                # Route Target,Format AS(4bytes):AN(2bytes)
                asn, an = struct.unpack('!IH', value_tmp)
                ext_community.append((bgp_cons.BGP_EXT_COM_RT_2, '%s:%s' % (asn, an)))

            # Route Origin
            elif comm_code == bgp_cons.BGP_EXT_COM_RO_0:
                # Route Origin,Format AS(2bytes):AN(4bytes)
                asn, an = struct.unpack('!HI', value_tmp)
                ext_community.append((bgp_cons.BGP_EXT_COM_RO_0, '%s:%s' % (asn, an)))

            elif comm_code == bgp_cons.BGP_EXT_COM_RO_1:
                # Route Origin,Format IP address:AN(2bytes)
                ipv4 = str(netaddr.IPAddress(struct.unpack('!I', value_tmp[0:4])[0]))
                an = struct.unpack('!H', value_tmp[4:])[0]
                ext_community.append((bgp_cons.BGP_EXT_COM_RO_1, '%s:%s' % (ipv4, an)))

            elif comm_code == bgp_cons.BGP_EXT_COM_RO_2:
                # Route Origin,Format AS(2bytes):AN(4bytes)
                asn, an = struct.unpack('!IH', value_tmp)
                ext_community.append((bgp_cons.BGP_EXT_COM_RO_2, '%s:%s' % (asn, an)))

            else:
                ext_community.append((bgp_cons.BGP_EXT_COM_UNKNOW, repr(value_tmp)))
                LOG.warn('unknow bgp extended community, type=%s, value=%s', comm_code, repr(value_tmp))

            value = value[8:]

        return ext_community

    @classmethod
    def construct(cls, value):

        """
        Construct Extended Community attributes.
        :param value: value list like [('RT':4837:9929),('RT': 1239:9929)]
        """
        ext_community_hex = b''
        for item in value:
            # for Route Target
            if item[0] == bgp_cons.BGP_EXT_COM_RT_0:
                # Route Target, Format AS(2bytes):AN(4bytes)
                asn, an = item[1].split(':')
                ext_community_hex += struct.pack('!HHI', bgp_cons.BGP_EXT_COM_RT_0, int(asn), int(an))
            elif item[0] == bgp_cons.BGP_EXT_COM_RT_1:
                ip, an = item[1].split(':')
                ext_community_hex += struct.pack('!H', bgp_cons.BGP_EXT_COM_RT_1) + netaddr.IPAddress(ip).packed + \
                    struct.pack('!H', int(an))
            elif item[0] == bgp_cons.BGP_EXT_COM_RT_2:
                asn, an = item[1].split(':')
                ext_community_hex += struct.pack('!HIH', bgp_cons.BGP_EXT_COM_RT_2, int(asn), int(an))

            # for Route Origin
            elif item[0] == bgp_cons.BGP_EXT_COM_RO_0:
                asn, an = item[1].split(':')
                ext_community_hex += struct.pack('!HHI', bgp_cons.BGP_EXT_COM_RO_0, int(asn), int(an))
            elif item[0] == bgp_cons.BGP_EXT_COM_RO_1:
                ip, an = item[1].split(':')
                ext_community_hex += struct.pack('!H', bgp_cons.BGP_EXT_COM_RO_1) + netaddr.IPAddress(ip).packed + \
                    struct.pack('!H', int(an))
            elif item[0] == bgp_cons.BGP_EXT_COM_RO_2:
                asn, an = item[1].split(':')
                ext_community_hex += struct.pack('!HIH', bgp_cons.BGP_EXT_COM_RO_2, int(asn), int(an))

            else:
                LOG.warn('unknow bgp extended community for construct, type=%s, value=%s', item[0], item[1])

        if ext_community_hex:
            return struct.pack('!B', cls.FLAG) + struct.pack('!B', cls.ID) \
                + struct.pack('!B', len(ext_community_hex)) + ext_community_hex
        else:
            LOG.error('construct error, value=%s' % value)
            return None



# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sys
import os
import logging

from oslo.config import cfg
from twisted.internet import reactor
from twisted.web.server import Site

from yabgp import version, log
from yabgp.core.factory import BGPPeering
from yabgp.config import get_bgp_config
from yabgp.common import constants as bgp_cons
from yabgp.api.app import app
from yabgp.channel.config import rabbit_mq
from yabgp.channel.factory import PikaFactory
from yabgp.db.config import database_options
from yabgp.db.mongodb import MongoApi


log.early_init_log(logging.DEBUG)

CONF = cfg.CONF

LOG = logging.getLogger(__name__)


def check_running_mode():

    if not CONF.standalone:
        # not standalone?
        CONF.register_opts(rabbit_mq, group='rabbit_mq')
        CONF.register_opts(database_options, group='database')


def check_msg_config():

    LOG.info('Check configurations about message process')
    if CONF.message.write_disk:
        if not os.path.exists(CONF.message.write_dir):
            os.makedirs(CONF.message.write_dir)
            LOG.info('Create dir %s', CONF.message.write_dir)
        CONF.message.write_msg_max_size = CONF.message.write_msg_max_size * 1024 * 1024


def prepare_twisted_service():
    LOG.info('Prepare twisted services')
    # check all peers
    all_peers = {}

    # check running mode
    if not CONF.standalone:
        # rabbitmq factory
        rabbit_mq_factory = PikaFactory(
            host=CONF.rabbit_mq.rabbit_host,
            port=CONF.rabbit_mq.rabbit_port,
            userid=CONF.rabbit_mq.rabbit_userid,
            password=CONF.rabbit_mq.rabbit_password
        )
        rabbit_mq_factory.connect()
        # mongodb connection
        mongo_connection = MongoApi(
            connection_url=CONF.database.connection,
            db_name=CONF.database.dbname,
            use_replica=CONF.database.use_replica,
            replica_name=CONF.database.replica_name,
            read_preference=CONF.database.read_preference,
            write_concern=CONF.database.write_concern,
            w_timeout=CONF.database.write_concern_timeout
        )
    else:
        rabbit_mq_factory = None
        mongo_connection = None
    for peer in CONF.bgp.running_config:
        LOG.info('Get peer %s configuration', peer)
        if CONF.message.write_disk:
            msg_file_path_for_peer = os.path.join(
                CONF.message.write_dir,
                peer.lower()
            )
            if not os.path.exists(msg_file_path_for_peer):
                os.makedirs(msg_file_path_for_peer)
                LOG.info('Create dir %s for peer %s', msg_file_path_for_peer, peer)
            LOG.info('BGP message file path is %s', msg_file_path_for_peer)
        else:
            msg_file_path_for_peer = None
        LOG.info('Create BGPPeering instance')
        afi_safi_list = [bgp_cons.AFI_SAFI_STR_DICT[afi_safi]
                         for afi_safi in CONF.bgp.running_config[peer]['afi_safi']]
        CONF.bgp.running_config[peer]['afi_safi'] = afi_safi_list
        CONF.bgp.running_config[peer]['capability']['local']['afi_safi'] = afi_safi_list
        bgp_peering = BGPPeering(
            myasn=CONF.bgp.running_config[peer]['local_as'],
            myaddr=CONF.bgp.running_config[peer]['local_addr'],
            peerasn=CONF.bgp.running_config[peer]['remote_as'],
            peeraddr=CONF.bgp.running_config[peer]['remote_addr'],
            afisafi=CONF.bgp.running_config[peer]['afi_safi'],
            msgpath=msg_file_path_for_peer,
            md5=CONF.bgp.running_config[peer]['md5'],
            channel=rabbit_mq_factory,
            mongo_conn=mongo_connection
        )
        all_peers[peer] = bgp_peering
        CONF.bgp.running_config[peer]['factory'] = bgp_peering

    # Starting api server
    if sys.version_info[0] == 2:
        from twisted.web.wsgi import WSGIResource
        LOG.info("Prepare RESTAPI service")
        resource = WSGIResource(reactor, reactor.getThreadPool(), app)
        site = Site(resource)
        try:
            reactor.listenTCP(CONF.rest.bind_port, site, interface=CONF.rest.bind_host)
            LOG.info("serving RESTAPI on http://%s:%s", CONF.rest.bind_host, CONF.rest.bind_port)
        except Exception as e:
            LOG.error(e, exc_info=True)
            sys.exit()

    for peer in all_peers:
        LOG.info('start peer, peer address=%s', peer)
        all_peers[peer].automatic_start()
    reactor.run()


def prepare_service(args=None):
    try:
        CONF(args=args, project='yabgp', version=version,
             default_config_files=['/etc/yabgp/yabgp.ini'])
    except cfg.ConfigFilesNotFoundError:
        CONF(args=args, project='yabgp', version=version)

    check_running_mode()
    log.init_log()
    LOG.info('Log (Re)opened.')
    LOG.info("Configuration:")
    cfg.CONF.log_opt_values(LOG, logging.INFO)
    try:
        get_bgp_config()
        check_msg_config()
    except Exception as e:
        LOG.error(e, exc_info=True)
        sys.exit()
    prepare_twisted_service()



# coding=utf-8
# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""IPv4 MPLS VPN NLRI
"""

import struct
import logging
import binascii

import netaddr

from yabgp.common import constants as bgp_cons

LOG = logging.getLogger(__name__)


class IPv4MPLSVPN(object):

    """
    IPv4 MPLS VPN NLRI

    +---------------------------+
    | Length (1 octet)          |
    +---------------------------+
    | Label  (3 octet)          |
    +---------------------------+
    |...........................|
    +---------------------------+
    | Prefix (variable)         |
    +---------------------------+
    a) Length: The Length field indicates the length, in bits, of the address prefix.
    b) Label: (24 bits) Carries one or more labels in a stack, although a BGP update
    has only one label. This field carries the following parts of the MPLS shim header:
        Label Value–—20 Bits
        Experimental bits—3 Bits
        Bottom of stack bit—1 bit
    c) Prefix: different coding way according to different SAFI
    Route Distinguisher (8 bytes) plus IPv4  prefix (32 bits) or IPv6 prefix(128 bits)
    rd (Route Distinguisher) structure (RFC 4364)
    """

    @classmethod
    def parse(cls, value):
        """
        parse nlri
        :param value: the raw hex nlri value
        :return: nlri list
        """
        nlri_list = []
        while value:
            nlri_dict = {}
            # for python2 and python3
            if isinstance(value[0], int):
                prefix_bit_len = value[0]
            else:
                prefix_bit_len = ord(value[0])
            if prefix_bit_len % 8 == 0:
                prefix_byte_len = int(prefix_bit_len / 8)
            else:
                prefix_byte_len = int(prefix_bit_len / 8) + 1

            label_int = int(binascii.b2a_hex(value[1:4]), 16)
            nlri_dict['label'] = (label_int >> 4, label_int & 1)

            rd = value[4:12]
            rd_type = struct.unpack('!H', rd[0:2])[0]
            rd_value = rd[2:8]
            nlri_dict['rd_type'] = rd_type
            if rd_type == bgp_cons.BGP_ROUTE_DISTINGUISHER_TYPE_0:
                asn, an = struct.unpack('!HI', rd_value)
                nlri_dict['rd'] = '%s:%s' % (asn, an)
            elif rd_type == bgp_cons.BGP_ROUTE_DISTINGUISHER_TYPE_1:
                ip = str(netaddr.IPAddress(struct.unpack('!I', rd_value[0:4])[0]))
                an = struct.unpack('!H', rd_value[4:6])[0]
                nlri_dict['rd'] = '%s:%s' % (ip, an)
            elif rd_type == bgp_cons.BGP_ROUTE_DISTINGUISHER_TYPE_2:
                asn, an = struct.unpack('!IH', rd_value)
                nlri_dict['rd'] = '%s:%s' % (asn, an)
            else:
                LOG.warning('unknow rd type for nlri, type=%s' % rd_type)
                nlri_dict['rd'] = '%s:%s' % (0, 0)
            prefix = value[12:prefix_byte_len + 1]
            if len(prefix) < 4:
                prefix += '\x00' * (4 - len(prefix))
            nlri_dict['str'] = str(netaddr.IPAddress(struct.unpack('!I', prefix)[0])) +\
                '/%s' % (prefix_bit_len - 88)
            value = value[prefix_byte_len + 1:]
            nlri_list.append(nlri_dict)
        return nlri_list

    @classmethod
    def contruct(cls, value):
        pass



# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

""" Test Aggregator attribute """

import unittest

from yabgp.common import exception as excep
from yabgp.common.constants import ERR_MSG_UPDATE_ATTR_LEN
from yabgp.message.attribute.aggregator import Aggregator


class TestAggregator(unittest.TestCase):
    def test_parse(self):

        # 4 bytes asn
        aggregator = Aggregator.parse(value=b'\x00\x00\x70\xd5\x3e\xe7\xff\x79',
                                      asn4=True)
        self.assertEqual((28885, '62.231.255.121'), aggregator)

        # 2 bytes asn
        aggregator = Aggregator.parse(value=b'\x70\xd5\x3e\xe7\xff\x79',
                                      asn4=False)
        self.assertEqual((28885, '62.231.255.121'), aggregator)

        # invalid attr len
        self.assertRaises(excep.UpdateMessageError, Aggregator.parse,
                          b'\x70\xd5\x3e\xe7\xff\x79',
                          True)
        try:
            Aggregator.parse(value=b'\x70\xd5\x3e\xe7\xff\x79',
                             asn4=True)
        except excep.UpdateMessageError as e:
            self.assertEqual(ERR_MSG_UPDATE_ATTR_LEN, e.sub_error)

    def test_construct(self):

        aggregator = Aggregator.construct(value=(28885, '62.231.255.121'))
        self.assertEqual(b'\xc0\x07\x06\x70\xd5\x3e\xe7\xff\x79', aggregator)


if __name__ == '__main__':
    unittest.main()



# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""BGP Attribute MP_UNREACH_NLRI
"""

from yabgp.message.attribute import Attribute
from yabgp.message.attribute import AttributeFlag
from yabgp.message.attribute import AttributeID


class MpUnReachNLRI(Attribute):

    """
        This is an optional non-transitive attribute that can be used for the
    purpose of withdrawing multiple unfeasible routes from service.
        An UPDATE message that contains the MP_UNREACH_NLRI is not required
    to carry any other path attributes.

        MP_UNREACH_NLRI coding information
        +---------------------------------------------------------+
        | Address Family Identifier (2 octets)                    |
        +---------------------------------------------------------+
        | Subsequent Address Family Identifier (1 octet)          |
        +---------------------------------------------------------+
        | Withdrawn Routes (variable)                             |
        +---------------------------------------------------------+
    """

    ID = AttributeID.MP_UNREACH_NLRI
    FLAG = AttributeFlag.OPTIONAL

    @classmethod
    def parse(cls, value):
        pass

    @classmethod
    def construct(cls, value):
        pass



# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

""" logging handler. Reference from https://github.com/osrg/ryu/blob/master/ryu/log.py
"""

from __future__ import print_function
import inspect
import logging
import logging.config
import logging.handlers
import os
import sys
if sys.version_info[0] == 2:
    import ConfigParser
elif sys.version_info[0] == 3:
    from configparser import ConfigParser

from oslo.config import cfg


CONF = cfg.CONF

CONF.register_cli_opts([
    cfg.BoolOpt('verbose', default=False, help='show debug output'),
    cfg.BoolOpt('use-stderr', default=True, help='log to standard error'),
    cfg.StrOpt('log-dir', default=None, help='log file directory'),
    cfg.StrOpt('log-file', default=None, help='log file name'),
    cfg.StrOpt('log-file-mode', default='0644',
               help='default log file permission'),
    cfg.StrOpt('log-config-file', default=None,
               help='Path to a logging config file to use')
])

DEBUG_LOG_FORMAT = '%(asctime)s.%(msecs)03d %(process)d %(levelname)s %(name)s ' \
                   '%(funcName)s %(lineno)d [-] %(message)s'
INFOR_LOG_FORMAT = '%(asctime)s.%(msecs)03d %(process)d %(levelname)s %(name)s [-] %(message)s'
_EARLY_LOG_HANDLER = None


def early_init_log(level=None):
    global _EARLY_LOG_HANDLER
    _EARLY_LOG_HANDLER = logging.StreamHandler(sys.stderr)

    log = logging.getLogger()
    log.addHandler(_EARLY_LOG_HANDLER)
    if level is not None:
        log.setLevel(level)


def _get_log_file():
    if CONF.log_file:
        return CONF.log_file
    if CONF.log_dir:
        return os.path.join(CONF.log_dir,
                            os.path.basename(inspect.stack()[-1][1])) + '.log'
    return None


def _set_log_format(handers, _format):
    for hander in handers:
        hander.setFormatter(logging.Formatter(_format))


def init_log():
    global _EARLY_LOG_HANDLER

    log = logging.getLogger()
    if CONF.log_config_file:
        try:
            logging.config.fileConfig(CONF.log_config_file,
                                      disable_existing_loggers=False)
            if CONF.verbose:
                log.setLevel(logging.DEBUG)
                for handler in log.handlers:
                    handler.setFormatter(logging.Formatter(DEBUG_LOG_FORMAT))
        except ConfigParser.Error as e:
            print('Failed to parse %s: %s' % (CONF.log_config_file, e),
                  file=sys.stderr)
            sys.exit(2)
        return

    if CONF.use_stderr:
        log.addHandler(logging.StreamHandler(sys.stderr))
    if _EARLY_LOG_HANDLER is not None:
        log.removeHandler(_EARLY_LOG_HANDLER)
        _EARLY_LOG_HANDLER = None

    log_file = _get_log_file()
    if log_file is not None:
        log.addHandler(logging.handlers.RotatingFileHandler(
            log_file, maxBytes=5 * 1024 * 1024, backupCount=5))
        mode = int(CONF.log_file_mode, 8)
        os.chmod(log_file, mode)
        for handler in log.handlers:
            handler.setFormatter(logging.Formatter(INFOR_LOG_FORMAT))

    if CONF.verbose:
        log.setLevel(logging.DEBUG)
        for handler in log.handlers:
            handler.setFormatter(logging.Formatter(DEBUG_LOG_FORMAT))
    else:
        log.setLevel(logging.INFO)
        _set_log_format(log.handlers, INFOR_LOG_FORMAT)



# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import struct

from yabgp.message.attribute import Attribute
from yabgp.message.attribute import AttributeFlag
from yabgp.message.attribute import AttributeID
from yabgp.common import constants as bgp_cons
from yabgp.common import exception as excep


class Origin(Attribute):
    """
        ORIGIN is a well-known mandatory attribute that defines the
    origin of the path information. The data octet can assume
    the following values:
    Value       Meaning
        0        IGP  -  Network Layer Reachability Information
                  is interior to the originating AS
        1        EGP - Network Layer Reachability Information
                  learned via the EGP protocol [RFC904]
        2        INCOMPLETE - Network Layer Reachability
                 Information learned by some other means
    """
    ID = AttributeID.ORIGIN
    FLAG = AttributeFlag.TRANSITIVE
    MULTIPLE = False

    IGP = 0x00
    EGP = 0x01
    INCOMPLETE = 0x02

    @classmethod
    def parse(cls, value):
        """
        Error process:
        (1) the falgs of the ORIGIN attribute must be "well-know,transitive"
        (2) If the ORIGIN attribute has an undefined value, then the Error Sub-code
        MUST be set to Invalid Origin Attribute. The Data field MUST contain the
        unrecognized attribute (type,length, and vlaue)
        :param value: raw binary value
        """
        orgin = struct.unpack('!B', value)[0]
        if orgin not in [cls.IGP, cls.EGP, cls.INCOMPLETE]:
            raise excep.UpdateMessageError(
                sub_error=bgp_cons.ERR_MSG_UPDATE_INVALID_ORIGIN,
                data=value)
        return orgin

    @classmethod
    def construct(cls, value):

        """
        construct a origin attribute
        :param value: 0,1,2
        """
        if value not in [cls.IGP, cls.EGP, cls.INCOMPLETE]:
            raise excep.UpdateMessageError(
                sub_error=bgp_cons.ERR_MSG_UPDATE_INVALID_ORIGIN,
                data='')
        length = 1
        return struct.pack('!B', cls.FLAG) + struct.pack('!B', cls.ID) \
            + struct.pack('!B', length) + struct.pack('!B', value)



# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""RFC 2858 subsequent address family numbers"""

SAFNUM_UNICAST = 1
SAFNUM_MULCAST = 2
SAFNUM_UNIMULC = 3
SAFNUM_MPLS_LABEL = 4  # rfc3107
SAFNUM_MCAST_VPN = 5  # draft-ietf-l3vpn-2547bis-mcast-bgp-08.txt
SAFNUM_ENCAPSULATION = 7  # rfc5512
SAFNUM_TUNNEL = 64  # draft-nalawade-kapoor-tunnel-safi-02.txt
SAFNUM_VPLS = 65
SAFNUM_MDT = 66  # rfc6037
SAFNUM_BGPLS = 71
SAFNUM_LAB_VPNUNICAST = 128  # Draft-rosen-rfc2547bis-03
SAFNUM_LAB_VPNMULCAST = 129
SAFNUM_LAB_VPNUNIMULC = 130
SAFNUM_ROUTE_TARGET = 132  # RFC 4684 Constrained Route Distribution for BGP/MPLS IP VPN
SAFNUM_FSPEC_RULE = 133  # RFC 5575 BGP flow spec SAFI
SAFNUM_FSPEC_VPN_RULE = 134  # RFC 5575 BGP flow spec SAFI VPN



# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import struct

from yabgp.message.attribute import Attribute
from yabgp.message.attribute import AttributeID
from yabgp.message.attribute import AttributeFlag
from yabgp.common import exception as excep
from yabgp.common import constants as bgp_cons


class ASPath(Attribute):
    """
        AS_PATH is a well-known mandatory attribute that is composed
    of a sequence of AS path segments. Each AS path segment is
    represented by a triple <path segment type, path segment
    length, path segment value>.
        The path segment type is a 1-octet length field with the
    following values defined:
    Value     Segment Type
        1       AS_SET: unordered set of ASes a route in the
                UPDATE message has traversed
        2       AS_SEQUENCE: ordered set of ASes a route in
                the UPDATE message has traversed
        3       AS_CONFED_SEQUENCE: ordered set of Member Autonomous
                Systems in the local confederation that the UPDATE message
                has traversed
        4       AS_CONFED_SET: unordered set of Member Autonomous Systems
                in the local confederation that the UPDATE message has
                traversed
        The path segment length is a 1-octet length field,
    containing the number of ASes (not the number of octets) in
    the path segment value field.
        The path segment value field contains one or more AS
    numbers, each encoded as a 2-octet length field.
    """
    AS_SET = 0x01
    AS_SEQUENCE = 0x02
    AS_CONFED_SEQUENCE = 0x03
    AS_CONFED_SET = 0x04

    ID = AttributeID.AS_PATH
    FLAG = AttributeFlag.TRANSITIVE
    MULTIPLE = False

    @classmethod
    def parse(cls, value, asn4=False):

        """
        Parse AS PATH attributes.
        :param asn4: 4 bytes asn or not
        :param value: raw binary balue
        """
        # Loop over all path segments
        aspath = []
        while len(value) > 0:
            seg_type, length = struct.unpack('!BB', value[:2])
            if seg_type not in [cls.AS_SET, cls.AS_SEQUENCE, cls.AS_CONFED_SEQUENCE, cls.AS_CONFED_SET]:
                raise excep.UpdateMessageError(
                    sub_error=bgp_cons.ERR_MSG_UPDATE_MALFORMED_ASPATH,
                    data=repr(value))
            try:
                if asn4:
                    segment = list(struct.unpack('!%dI' % length, value[2:2 + length * 4]))
                    value = value[2 + length * 4:]

                else:
                    segment = list(struct.unpack('!%dH' % length, value[2:2 + length * 2]))
                    value = value[2 + length * 2:]
            except Exception:
                raise excep.UpdateMessageError(
                    sub_error=bgp_cons.ERR_MSG_UPDATE_ATTR_LEN,
                    data='')
            aspath.append((seg_type, segment))
        return aspath

    @classmethod
    def construct(cls, value, asn4=False):

        """
        Construct AS PATH.
        :param asn4: 4byte asn
        :param value:
        """
        # value example
        # [(2, [3257, 31027, 34848, 21465])], or [(3, [64606]), (2, [64624, 65515])]

        as_path_raw = b''
        for segment in value:
            as_seg_raw = b''
            seg_type = segment[0]
            as_path_list = segment[1]
            if seg_type not in [cls.AS_SET, cls.AS_SEQUENCE, cls.AS_CONFED_SET, cls.AS_CONFED_SEQUENCE]:
                assert excep.UpdateMessageError(
                    sub_error=bgp_cons.ERR_MSG_UPDATE_MALFORMED_ASPATH,
                    data='')

            if asn4:
                #  4 bytes asn encode
                as_count = 0
                for asn in as_path_list:
                    as_count += 1
                    as_seg_raw += struct.pack('!I', asn)
            else:
                # 2 bytes asn encode
                as_count = 0
                for asn in as_path_list:
                    as_count += 1
                    as_seg_raw += struct.pack('!H', asn)

            as_path_raw += struct.pack('!B', seg_type) + struct.pack('!B', as_count) + as_seg_raw

        flags = cls.FLAG
        if len(as_path_raw) > 255:
            flags += AttributeFlag.EXTENDED_LENGTH
            return struct.pack('!B', flags) + struct.pack('!B', cls.ID) \
                + struct.pack('!H', len(as_path_raw)) + as_path_raw
        else:
            return struct.pack('!B', flags) + struct.pack('!B', cls.ID) \
                + struct.pack('!B', len(as_path_raw)) + as_path_raw



# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import binascii
import struct

import netaddr

from yabgp.message.attribute import Attribute
from yabgp.message.attribute import AttributeFlag
from yabgp.message.attribute import AttributeID
from yabgp.common import exception as excep
from yabgp.common import constants as bgp_cons


class OriginatorID(Attribute):
    """
        ORIGINATOR_ID is a new optional, non-transitive BGP attribute of Type
    code 9. This attribute is 4 bytes long and it will be created by an
    RR in reflecting a route. This attribute will carry the BGP
    Identifier of the originator of the route in the local AS. A BGP
    speaker SHOULD NOT create an ORIGINATOR_ID attribute if one already
    exists. A router that recognizes the ORIGINATOR_ID attribute SHOULD
    ignore a route received with its BGP Identifier as the ORIGINATOR_ID.
    (RFC 4456 Page 6)
    """

    ID = AttributeID.ORIGINATOR_ID
    FLAG = AttributeFlag.OPTIONAL

    @classmethod
    def parse(cls, value):

        """
        Parse originator id
        :param value:
        """
        if len(value) != 4:
            raise excep.UpdateMessageError(
                sub_error=bgp_cons.ERR_MSG_UPDATE_ATTR_LEN,
                data=value)
        return str(netaddr.IPAddress(int(binascii.b2a_hex(value[0:4]), 16)))

    @classmethod
    def construct(cls, value):

        """
        construct a ORIGINATOR_ID path attribute
        :param value: ipv4 format string
        """
        try:
            return struct.pack('!B', cls.FLAG) + struct.pack('!B', cls.ID) \
                + struct.pack('!B', 4) + netaddr.IPAddress(value).packed
        except Exception:
            raise excep.UpdateMessageError(
                sub_error=bgp_cons.ERR_MSG_UPDATE_ATTR_LEN,
                data=value)



# Copyright (c) 2007 by Mark Bergsma <mark@nedworks.org>
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Twisted Factory, BGP implementation.
"""

import time
import os
import logging
import socket
import platform
import struct
import sys

import netaddr
from twisted.internet import protocol
from twisted.internet import reactor
from oslo.config import cfg

from yabgp.core.protocol import BGP
from yabgp.core.fsm import FSM
from yabgp.common import constants as bgp_cons
from yabgp.common.afn import AFNUM_INET

LOG = logging.getLogger(__name__)

CONF = cfg.CONF


class BGPFactory(protocol.Factory):
    """Base factory for creating BGP protocol instances."""

    protocol = BGP
    FSM = FSM

    def buildProtocol(self, addr):
        """Builds a BGPProtocol instance.

        :param addr : address used for buliding protocol.
        """
        return protocol.Factory.buildProtocol(self, addr)

    def startedConnecting(self, connector):
        """Called when a connection attempt has been initiated.

        :param connector : Twisted connector
        """
        pass

    def clientConnectionLost(self, connector, reason):
        """ Called when a TCP client connection was lost.

        :param connector : Twisted connector
        :param reason : connection faied reason.
        """
        LOG.info("Client connection lost:%s", reason.getErrorMessage())


class BGPPeering(BGPFactory):
    """Class managing a BGP session with a peer.
    One connection, One BGPPeering class.
    """

    def __init__(self, myasn=None, myaddr=None, peerasn=None, peeraddr=None,
                 msgpath=None, afisafi=None, md5=None, channel=None, mongo_conn=None):
        """Initial a BGPPeering instance.

        :param myasn: local bgp as number.
        :param myaddr: local ip address.
        :param peerasn: remote bgp peer as number
        :param peeraddr: remote peer ip address.
        :param msgpath: the path to store bgp message file.
        :param afisafi: afi and safi
        :param md5: TCP md5 string
        :param channel: rabbitmq channel
        :param mongo_conn: mongodb connection
        """
        LOG.info('Init BGPPeering for peer %s', peeraddr)
        self.my_asn = myasn
        self.my_addr = myaddr
        self.peer_addr = peeraddr
        self.peer_id = None
        self.bgp_id = None
        self.peer_asn = peerasn
        self.afi_safi = afisafi
        self.md5 = md5
        self.channel = channel
        self.mongo_conn = mongo_conn

        self.status = False
        self.fsm = BGPFactory.FSM(self)

        # reference to the BGPProtocol instance in ESTAB state
        self.estab_protocol = None
        self.msg_path = msgpath
        if msgpath:
            self.msg_path = msgpath + '/msg/'
            if not os.path.exists(self.msg_path):
                os.makedirs(self.msg_path)
            self.msg_file_name = "%s.msg" % time.time()
            self.msg_seq = self.get_last_seq()
            self.msg_file = open(os.path.join(self.msg_path, self.msg_file_name), 'a')
            self.msg_file.flush()
            LOG.info('BGP message file %s', self.msg_file_name)
            LOG.info('The last bgp message seq number is %s', self.msg_seq)

    def buildProtocol(self, addr):

        """Builds a BGP protocol instance

        :param addr: IP address used for building protocol.
        """
        LOG.info("[%s]Building a new BGP protocol instance", self.peer_addr)

        p = BGPFactory.buildProtocol(self, addr)
        if p is not None:
            self._initProtocol(p, addr)
        return p

    def _initProtocol(self, protocol, addr):

        """Initializes a BGPProtocol instance

        :param protocol: twisted Protocol
        :param addr: ip address
        """

        protocol.bgp_peering = self

        # Hand over the FSM
        protocol.fsm = self.fsm
        protocol.fsm.protocol = protocol

        if addr.port == bgp_cons.PORT:
            protocol.fsm.state = bgp_cons.ST_CONNECT
        else:
            protocol.fsm.state = bgp_cons.ST_ACTIVE

    def clientConnectionFailed(self, connector, reason):

        """Called when the outgoing connection failed.

        :param connector: Twisted connector
        :param reason: connection failed reason
        """

        error_msg = "[%s]Client connection failed: ", self.peer_addr, reason.getErrorMessage()
        if self.msg_path:
            self.write_msg(
                timestamp=time.time(),
                msg_type=0,
                msg=error_msg,
                afi_safi=(0, 0),
                flush=True
            )
        LOG.info(error_msg)
        # There is no protocol instance yet at this point.
        # Catch a possible NotificationException
        try:
            self.fsm.connection_failed()
        except Exception as e:
            LOG.info("[%s]Client connection failed: %s", self.peer_addr, e)

    def get_last_seq(self):

        """
        Get the last sequence number in the latest log file.
        """
        LOG.info('get the last bgp message seq for this peer')
        last_seq = 1
        # first get the last file
        file_list = os.listdir(self.msg_path)
        if not file_list:
            return last_seq
        file_list.sort()
        self.msg_file_name = file_list[-1]
        with open(self.msg_path + self.msg_file_name, 'rb') as fh:
            for line in fh:
                pass
            last = line
            last_seq = eval(last)[1]
        return last_seq

    def write_msg(self, timestamp, msg_type, msg, afi_safi, flush=False):
        """

        :param timestamp:
        :param msg_type:
        :param msg:
        :param afi_safi:
        :param flush:
        :return:
        """
        if self.msg_path:
            msg_record = [timestamp, self.msg_seq, msg_type, msg, afi_safi]
            self.msg_file.write(str(msg_record) + '\n')
            self.msg_seq += 1
            if flush:
                self.msg_file.flush()

    def flush_and_check_file_size(self):
        if self.msg_path:
            # Flush message log file
            self.msg_file.flush()

            # if the size of the msg file is bigger than 'max_msg_file_size',
            # then save as and re-open a new file.
            if os.path.getsize(self.msg_path + self.msg_file_name) > \
                    CONF.message.write_msg_max_size:
                self.msg_file.close()

                self.msg_file_name = "%s.msg", time.time()
                LOG.info('Open a new message file %s', self.msg_file_name)
                self.msg_file = open(self.msg_path + self.msg_file_name, 'ab')
                return True
        return False

    def automatic_start(self, idle_hold=False):

        """BGP AutomaticStart event (event 3)

        :param idle_hold: flag represents used Idle Hold
        """
        if self.fsm.state == bgp_cons.ST_IDLE:
            if self.fsm.automatic_start(idle_hold):
                self.status = True
                # Create outbound connection as a client
                self.connect()

    def manual_stop(self):

        """BGP ManualStop event (event 2) Returns a DeferredList that
        will fire once the connection(s) have closed"""

        for c in self.in_connections + self.out_connections:
            # Catch a possible NotificationSent exception
            try:
                c.fsm.manual_stop()
            except Exception as e:
                LOG.error(e)

    def connection_closed(self, pro, disconnect=False):
        """
        Called by FSM or Protocol when the BGP connection has been closed.

        :param pro: twisted protocol
        :param disconnect: the status of connection
        """

        LOG.info("[%s]Connection closed", self.peer_addr)
        if pro is not None:
            # Connection succeeded previously, protocol exists
            # Remove the protocol, if it exists
            if pro is self.estab_protocol:
                self.estab_protocol = None
                # self.fsm should still be valid and set to ST_IDLE
                self.fsm.state = bgp_cons.ST_IDLE

        if self.fsm.allow_automatic_start:
            self.automatic_start(idle_hold=True)

    def connect_retry(self):

        """Called by FSM when we should reattempt to connect.
        """
        self.connect()

    def set_peer_id(self, bgp_id):
        """
        Should be called when an Open message was received from a peer.
        Sets the BGP identifier of the peer if it wasn't set yet. If the
        new peer id is unequal to the existing one, CEASE all connections.
        s
        :param bgp_id: BGP ID
        """

        if self.peer_id is None:
            self.peer_id = bgp_id
            LOG.info('Set BGP peer id %s', bgp_id)
        elif self.peer_id != bgp_id:
            # Ouch, schizophrenia. The BGP id of the peer is unequal to
            # the ids of current and/or previous sessions. Close all
            # connections.
            self.peer_id = None

    def connect(self):

        """Initiates a TCP client connection to the peer. Should only be called from
        BGPPeering or FSM, otherwise use manualStart() instead.
        """
        # DEBUG
        LOG.info("(Re)connect to %s", self.peer_addr)

        if self.fsm.state != bgp_cons.ST_ESTABLISHED:

            connector = reactor.connectTCP(
                host=self.peer_addr,
                port=bgp_cons.PORT,
                factory=self,
                timeout=30,
                bindAddress=(self.my_addr, 0))
            if isinstance(self.md5, str) and self.md5:
                md5sig = self.get_tcp_md5sig(self.md5, self.peer_addr, bgp_cons.PORT)
                if md5sig:
                    connector.transport.getHandle().setsockopt(socket.IPPROTO_TCP, bgp_cons.TCP_MD5SIG, md5sig)
                else:
                    sys.exit()
            return True
        else:
            return False

    @staticmethod
    def get_tcp_md5sig(md5_str, host, port):

        os_type = platform.system()
        if os_type != 'Linux':
            LOG.error('YABGP has no MD5 support for %s', os_type)
            return None

        # address family
        if netaddr.IPAddress(host).version == 4:
            # ipv4 address
            afi = AFNUM_INET
        elif netaddr.IPAddress(host).version == 6:
            # ipv6 address
            # TODO support IPv6
            LOG.error("Does not support ipv6 address family")
            sys.exit()
        else:
            afi = None
            LOG.error('Peer address is not a valid ipv4/ipv6 address')
        try:
            n_port = socket.htons(port)
            if afi == AFNUM_INET:
                n_addr = socket.inet_pton(socket.AF_INET, host)
                tcp_md5sig = 'HH4s%dx2xH4x%ds' % (bgp_cons.SS_PADSIZE_IPV4, bgp_cons.TCP_MD5SIG_MAXKEYLEN)
                md5sig = struct.pack(tcp_md5sig, socket.AF_INET, n_port, n_addr, len(md5_str), md5_str.encode())
                return md5sig
            else:
                return None
        except socket.error as e:
            LOG.error('This linux machine does not support TCP_MD5SIG, you can not use MD5 (%s)', str(e))
            return None
